
/*
 *  selector is a simple command line utility for selection of strings
 *  with a dynamic pattern-matching.
 *
 *  Copyright (c) 2009, 2010, 2011, 2012 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of selector.
 *
 *  selector is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  selector is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with selector.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*

  To use it as a super-history-search for bash:
  selector --bash <(history)

*/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ncurses.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <regex.h>
#include <locale.h>
#include <getopt.h>
#include <limits.h>

#define VERSION "1.1.6"

#define BUFFER_SIZE 4096

/* Yeah, global variables! */

int nb_lines_max = 1000;
char pattern_separator = ';';
char label_separator = '\0';
int output_to_vt_buffer = 0;
int add_control_qs = 0;
int with_colors = 1;
int zsh_history = 0;
int bash_history = 0;
int inverse_order = 0;
int remove_duplicates = 0;
int use_regexp = 0;
int case_sensitive = 0;
char *title = 0;
int error_flash = 0;
int upper_caps_makes_case_sensitive = 0;
int show_long_lines = 0;
int show_hits = 0;

int attr_modeline, attr_focus_line, attr_error, attr_hits;

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if(!p && n != 0) {
    fprintf(stderr,
            "selector: can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/*********************************************************************/

void inject_into_tty_buffer(char *string, int add_control_qs) {
  struct termios oldtio, newtio;
  const char *k;
  const char control_q = '\021';
  tcgetattr(STDIN_FILENO, &oldtio);
  memset(&newtio, 0, sizeof(newtio));
  /* Set input mode (non-canonical, *no echo*,...) */
  tcsetattr(STDIN_FILENO, TCSANOW, &newtio);
  /* Put the selected string in the tty input buffer */
  for(k = string; *k; k++) {
    if(add_control_qs && !(*k >= ' ' && *k <= '~')) {
      /* Add ^Q to quote control characters */
      ioctl(STDIN_FILENO, TIOCSTI, &control_q);
    }
    ioctl(STDIN_FILENO, TIOCSTI, k);
  }
  /* Restore the old settings */
  tcsetattr(STDIN_FILENO, TCSANOW, &oldtio);
}

/*********************************************************************/

void str_to_positive_integers(char *string, int *values, int nb) {
  int current_value, gotone;
  char *s;
  int n;

  n = 0;
  current_value = 0;
  gotone = 0;
  s = string;

  while(1) {
    if(*s >= '0' && *s <= '9') {
      current_value = current_value * 10 + (int) (*s - '0');
      gotone = 1;
    } else if(*s == ',' || *s == '\0') {
      if(gotone) {
        if(n < nb) {
          values[n++] = current_value;
          if(*s == '\0') {
            if(n == nb) {
              return;
            } else {
              fprintf(stderr,
                      "selector: Missing value in `%s'.\n", string);
              exit(EXIT_FAILURE);
            }
          }
          current_value = 0;
          gotone = 0;
        } else {
          fprintf(stderr,
                  "selector: Too many values in `%s'.\n", string);
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr,
                "selector: Empty value in `%s'.\n", string);
        exit(EXIT_FAILURE);
      }
    } else {
      fprintf(stderr,
              "selector: Syntax error in `%s'.\n", string);
      exit(EXIT_FAILURE);
    }
    s++;
  }
}

void error_feedback() {
  if(error_flash) {
    flash();
  } else {
    beep();
  }
}

void usage(FILE *out) {

  fprintf(out, "Selector version %s (%s)\n", VERSION, UNAME);
  fprintf(out, "Written by Francois Fleuret <francois@fleuret.org>.\n");
  fprintf(out, "\n");
  fprintf(out, "Usage: selector [options] [<filename1> [<filename2> ...]]\n");
  fprintf(out, "\n");
  fprintf(out, " -h, --help\n");
  fprintf(out, "         show this help\n");
  fprintf(out, " -v, --inject-in-tty\n");
  fprintf(out, "         inject the selected line in the tty\n");
  fprintf(out, " -w, --add-control-qs\n");
  fprintf(out, "         quote control characters with ^Qs when using -v\n");
  fprintf(out, " -d, --remove-duplicates\n");
  fprintf(out, "         remove duplicated lines\n");
  fprintf(out, " -b, --remove-bash-prefix\n");
  fprintf(out, "         remove the bash history line prefix\n");
  fprintf(out, " -z, --remove-zsh-prefix\n");
  fprintf(out, "         remove the zsh history line prefix\n");
  fprintf(out, " -i, --revert-order\n");
  fprintf(out, "         invert the order of lines\n");
  fprintf(out, " -e, --regexp\n");
  fprintf(out, "         start in regexp mode\n");
  fprintf(out, " -a, --case-sensitive\n");
  fprintf(out, "         start in case sensitive mode\n");
  fprintf(out, " -j, --show-long-lines\n");
  fprintf(out, "         print three dots at the end of truncated lines\n");
  fprintf(out, " -y, --show-hits\n");
  fprintf(out, "         highlight the matching substrings\n");
  fprintf(out, " -u, --upper-case-makes-case-sensitive\n");
  fprintf(out, "         using an upper case character in the matching string makes\n");
  fprintf(out, "         the matching case-sensitive\n");
  fprintf(out, " -m, --monochrome\n");
  fprintf(out, "         monochrome mode\n");
  fprintf(out, " -q, --no-beep\n");
  fprintf(out, "         make a flash instead of a beep on an edition error\n");
  fprintf(out, " --bash\n");
  fprintf(out, "         setting for bash history search, same as -b -i -d -v -w -l ${HISTSIZE}\n");
  fprintf(out, " --      all following arguments are filenames\n");
  fprintf(out, " -t <title>, --title <title>\n");
  fprintf(out, "         add a title in the modeline\n");
  fprintf(out, " -c <colors>, --colors <colors>\n");
  fprintf(out, "         set the display colors with an argument of the form\n");
  fprintf(out, "         <fg_modeline>,<bg_modeline>,<fg_highlight>,<bg_highlight>\n");
  fprintf(out, " -o <output filename>, --output-file <output filename>\n");
  fprintf(out, "         set a file to write the selected line to\n");
  fprintf(out, " -s <pattern separator>, --pattern-separator <pattern separator>\n");
  fprintf(out, "         set the symbol to separate substrings in the pattern\n");
  fprintf(out, " -x <label separator>, --label-separator <label separator>\n");
  fprintf(out, "         set the symbol to terminate the label\n");
  fprintf(out, " -l <max number of lines>, --number-of-lines <max number of lines>\n");
  fprintf(out, "         set the maximum number of lines to take into account\n");
  fprintf(out, "\n");
}

/*********************************************************************/

/* A quick and dirty hash table */

#define MAGIC_HASH_MULTIPLIER 387433

/* The table itself stores indexes of the strings taken in a char**
   table. When a string is added, if it was already in the table, the
   new index replaces the previous one.  */

struct hash_table_t {
  int size;
  int *entries;
};

struct hash_table_t *new_hash_table(int size) {
  int k;
  struct hash_table_t *hash_table;

  hash_table = safe_malloc(sizeof(struct hash_table_t));

  hash_table->size = size;
  hash_table->entries = safe_malloc(hash_table->size * sizeof(int));

  for(k = 0; k < hash_table->size; k++) {
    hash_table->entries[k] = -1;
  }

  return hash_table;
}

void free_hash_table(struct hash_table_t *hash_table) {
  free(hash_table->entries);
  free(hash_table);
}

/* Adds new_string in the table, associated to new_index. If this
   string was not already in the table, returns -1. Otherwise, returns
   the previous index it had. */

int add_and_get_previous_index(struct hash_table_t *hash_table,
                               const char *new_string, int new_index,
                               char **strings) {

  unsigned int code = 0, start;
  int k;

  /* This is my recipe. I checked, it seems to work (as long as
     hash_table->size is not a multiple of MAGIC_HASH_MULTIPLIER that
     should be okay) */

  for(k = 0; new_string[k]; k++) {
    code = code * MAGIC_HASH_MULTIPLIER + (unsigned int) (new_string[k]);
  }

  code = code % hash_table->size;
  start = code;

  while(hash_table->entries[code] >= 0) {
    /* There is a string with that code */
    if(strcmp(new_string, strings[hash_table->entries[code]]) == 0) {
      /* It is the same string, we keep a copy of the stored index */
      int result = hash_table->entries[code];
      /* Put the new one */
      hash_table->entries[code] = new_index;
      /* And return the previous one */
      return result;
    }
    /* This collision was not the same string, let's move to the next
       in the table */
    code = (code + 1) % hash_table->size;
    /* We came back to our original code, which means that the table
       is full */
    if(code == start) {
      fprintf(stderr,
              "Full hash table (that should not happen)\n");
      exit(EXIT_FAILURE);
   }
  }

  /* This string was not already in there, store the index in the
     table and return -1 */

  hash_table->entries[code] = new_index;
  return -1;
}

/*********************************************************************
 A matcher matches either with a collection of substrings, or with a
 regexp */

struct matcher {
  regex_t preg;
  int regexp_error;
  int nb_patterns;
  int case_sensitive;
  char *splitted_patterns, **patterns;
};

/* Routine to add an interval to a sorted list of intervals
   extremities. Returns the resulting number of extremities.

   This routine is an effing nightmare */

int add_interval(int n, int *switches, int start, int end) {
  int f, g, k;

  if(start == end) { return n; }

  f = 0;
  while(f < n && switches[f] <= start) { f++; }
  g = f;
  while(g < n && switches[g] <= end) { g++; }

  if(f == n) {
    /* switches[n]   start  end  */
    /* XXXXXXXXXX|               */
    switches[f] = start;
    switches[f+1] = end;
    return n + 2;
  }

  if(f % 2) {

    if(g % 2) {
      /* switches[f-1]   start   switches[f]         switches[g-1]   end    switches[g] */
      /* |XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX|   ...   |XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX| */
      for(k = f; k < n; k++) { switches[k] = switches[k + (g - f)]; }
      return n - (g - f);
    } else {
      /* switches[f-1]   start   switches[f]         switches[g-1]   end    switches[g] */
      /* |XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX|   ...   XXXXXXXXXXXX|          |XXXXXXXXXX */
      switches[g - 1] = end;
      for(k = f; k < n; k++) { switches[k] = switches[k + ((g - 1) - f)]; }
      return n - ((g - 1) - f);
    }

  } else {

    if(f == g) {
      /* switches[f-1]   start  end   switches[f]  */
      /* XXXXXXXXXXXX|                |XXXXXXXXXX  */
      for(k = n - 1; k >= f; k--) {
        switches[k + 2] = switches[k];
      }
      switches[f] = start;
      switches[f + 1] = end;
      return n + 2;
    }

    if(g % 2) {
      /* switches[f-1]   start   switches[f]         switches[g-1]   end    switches[g] */
      /* XXXXXXXXXXXX|           |XXXXXXXXXX   ...   |XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX| */
      switches[f] = start;
      for(k = f + 1; k < n; k++) { switches[k] = switches[k + (g - (f + 1))]; }
      return n - (g - (f + 1));
    } else {
      /* switches[f-1]   start   switches[f]         switches[g-1]   end    switches[g] */
      /* XXXXXXXXXXXX|           |XXXXXXXXXX   ...   XXXXXXXXXXXX|          |XXXXXXXXXX */
      switches[f] = start;
      switches[g - 1] = end;
      for(k = f + 1; k < n; k++) { switches[k] = switches[k + ((g - 1) - (f + 1))]; }
      return n - ((g - 1) - (f + 1));
    }
  }
}

int match(struct matcher *matcher, char *string, int *nb_switches, int *switches) {
  int n;
  char *where;
  regmatch_t matches;

  if(nb_switches) { *nb_switches = 0; }

  if(matcher->nb_patterns >= 0) {
    if(matcher->case_sensitive) {
      for(n = 0; n < matcher->nb_patterns; n++) {
        if((where = strstr(string, matcher->patterns[n])) == 0) return 0;
        if(switches) {
          *nb_switches = add_interval(*nb_switches, switches,
                                      (int) (where - string),
                                      (int) (where - string) + strlen(matcher->patterns[n]));
        }
      }
    } else {
      for(n = 0; n < matcher->nb_patterns; n++) {
        if((where = strcasestr(string, matcher->patterns[n])) == 0) return 0;
        if(switches) {
          *nb_switches = add_interval(*nb_switches, switches,
                                      (int) (where - string),
                                      (int) (where - string) + strlen(matcher->patterns[n]));
        }
      }
    }
    return 1;
  } else {
    if(switches) {
      if(regexec(&matcher->preg, string, 1, &matches, 0) == 0) {
        *nb_switches = 2;
        switches[0] = matches.rm_so;
        switches[1] = matches.rm_eo;
        return 1;
      } else {
        return 0;
      }
    } else {
      return regexec(&matcher->preg, string, 0, 0, 0) == 0;
    }
  }
}

void free_matcher(struct matcher *matcher) {
  if(matcher->nb_patterns < 0) {
    if(!matcher->regexp_error) regfree(&matcher->preg);
  } else {
    free(matcher->splitted_patterns);
    free(matcher->patterns);
  }
}

void initialize_matcher(struct matcher *matcher,
                        int use_regexp, int case_sensitive,
                        const char *pattern) {
  const char *s;
  char *t, *last_pattern_start;
  int n;

  if(use_regexp) {
    matcher->case_sensitive = case_sensitive;
    matcher->nb_patterns = -1;
    matcher->regexp_error = regcomp(&matcher->preg, pattern,
                                    case_sensitive ? 0 : REG_ICASE);
  } else {
    matcher->regexp_error = 0;
    matcher->nb_patterns = 1;

    if(upper_caps_makes_case_sensitive) {
      for(s = pattern; *s && !case_sensitive; s++) {
        case_sensitive = (*s >= 'A' && *s <= 'Z');
      }
    }

    matcher->case_sensitive = case_sensitive;

    for(s = pattern; *s; s++) {
      if(*s == pattern_separator) {
        matcher->nb_patterns++;
      }
    }

    matcher->splitted_patterns =
      safe_malloc((strlen(pattern) + 1) * sizeof(char));

    matcher->patterns =
      safe_malloc(matcher->nb_patterns * sizeof(char *));

    strcpy(matcher->splitted_patterns, pattern);

    n = 0;
    last_pattern_start = matcher->splitted_patterns;
    for(t = matcher->splitted_patterns; n < matcher->nb_patterns; t++) {
      if(*t == pattern_separator || *t == '\0') {
        *t = '\0';
        matcher->patterns[n++] = last_pattern_start;
        last_pattern_start = t + 1;
      }
    }
  }
}

/*********************************************************************
 Buffer edition */

void delete_char(char *buffer, int *position) {
  if(buffer[*position]) {
    int c = *position;
    while(c < BUFFER_SIZE && buffer[c]) {
      buffer[c] = buffer[c+1];
      c++;
    }
  } else error_feedback();
}

void backspace_char(char *buffer, int *position) {
  if(*position > 0) {
    if(buffer[*position]) {
      int c = *position - 1;
      while(buffer[c]) {
        buffer[c] = buffer[c+1];
        c++;
      }
    } else {
      buffer[*position - 1] = '\0';
    }

    (*position)--;
  } else error_feedback();
}

void insert_char(char *buffer, int *position, char character) {
  if(strlen(buffer) < BUFFER_SIZE - 1) {
    int c = *position;
    char t = buffer[c], u;
    while(t) {
      c++;
      u = buffer[c];
      buffer[c] = t;
      t = u;
    }
    c++;
    buffer[c] = '\0';
    buffer[(*position)++] = character;
  } else error_feedback();
}

void kill_before_cursor(char *buffer, int *position) {
  int s = 0;
  while(buffer[*position + s]) {
    buffer[s] = buffer[*position + s];
    s++;
  }
  buffer[s] = '\0';
  *position = 0;
}

void kill_after_cursor(char *buffer, int *position) {
  buffer[*position] = '\0';
}

/*********************************************************************/

int previous_visible(int current_line, char **lines, struct matcher *matcher) {
  int line = current_line - 1;
  while(line >= 0 && !match(matcher, lines[line], 0, 0)) line--;
  return line;
}

int next_visible(int current_line, int nb_lines, char **lines,
                 struct matcher *matcher) {
  int line = current_line + 1;
  while(line < nb_lines && !match(matcher, lines[line], 0, 0)) line++;

  if(line < nb_lines)
    return line;
  else
    return -1;
}

/*********************************************************************/

void print_string_with_switches(char *buffer, int line_width,
                                int console_width,
                                int nb_patterns, int *switches) {
  int w, current = 0, next;
  if(switches) {
    for(w = 0; w < nb_patterns && switches[2 * w] < line_width; w++) {
      if(switches[2 * w] < switches[2 * w + 1]) {
        next = switches[2 * w];
        if(next > line_width) { next = line_width; }
        if(next > current) { addnstr(buffer + current,  next - current); }
        attron(attr_hits);
        current = next;
        next = switches[2 * w + 1];
        if(next > line_width) { next = line_width; }
        if(next > current) { addnstr(buffer + current,  next - current); }
        attroff(attr_hits);
        current = next;
      }
    }
    if(current < line_width) {
      addnstr(buffer + current, console_width - current);
    }
  } else {
    addnstr(buffer, console_width);
  }
}

/* The line highlighted is the first one matching the matcher in that
   order: (1) current_focus_line after motion, if it does not match,
   then (2) the first with a greater index, if none matches, then (3)
   the first with a lesser index.

   The index of the line actually shown highlighted is written in
   displayed_focus_line (it can be -1 if no line at all matches the
   matcher)

   If there is a motion and a line is actually shown highlighted, its
   value is written in current_focus_line. */

void update_screen(int *current_focus_line, int *displayed_focus_line,
                   int motion,
                   int nb_lines, char **lines,
                   int cursor_position,
                   char *pattern) {
  int *switches;
  char buffer[BUFFER_SIZE];
  struct matcher matcher;
  int k, l, m;
  int console_width, console_height;
  int nb_printed_lines = 0;
  int cursor_x;
  int nb_switches;

  initialize_matcher(&matcher, use_regexp, case_sensitive, pattern);

  if(show_hits) {
    if(matcher.nb_patterns >= 0) {
      switches = safe_malloc(sizeof(int) * matcher.nb_patterns * 2);
    } else {
      switches = safe_malloc(sizeof(int) * 2);
    }
  } else {
    switches = 0;
  }

  console_width = getmaxx(stdscr);
  console_height = getmaxy(stdscr);

  use_default_colors();

  /* Add an empty line where we will print the modeline at the end */

  addstr("\n");

  /* If the regexp is erroneous, print a message saying so */

  if(matcher.regexp_error) {
    attron(attr_error);
    addnstr("Regexp syntax error", console_width);
    attroff(attr_error);
  }

  /* Else, and we do have lines to select from, find a visible line. */

  else if(nb_lines > 0) {
    int new_focus_line;
    if(match(&matcher, lines[*current_focus_line], 0, 0)) {
      new_focus_line = *current_focus_line;
    } else {
      new_focus_line = next_visible(*current_focus_line, nb_lines, lines,
                                    &matcher);
      if(new_focus_line < 0) {
        new_focus_line = previous_visible(*current_focus_line, lines, &matcher);
      }
    }

    /* If we found a visible line and we should move, let's move */

    if(new_focus_line >= 0 && motion != 0) {
      int l = new_focus_line;
      if(motion > 0) {
        /* We want to go down, let's find the first visible line below */
        for(m = 0; l >= 0 && m < motion; m++) {
          l = next_visible(l, nb_lines, lines, &matcher);
          if(l >= 0) {
            new_focus_line = l;
          }
        }
      } else {
        /* We want to go up, let's find the first visible line above */
        for(m = 0; l >= 0 && m < -motion; m++) {
          l = previous_visible(l, lines, &matcher);
          if(l >= 0) {
            new_focus_line = l;
          }
        }
      }
    }

    /* Here new_focus_line is either a line number matching the
       pattern, or -1 */

    if(new_focus_line >= 0) {

      int first_line = new_focus_line, last_line = new_focus_line;
      int nb_match = 1;

      /* We find the first and last lines to show, so that the total
         of visible lines between them (them included) is
         console_height-1 */

      while(nb_match < console_height-1 &&
            (first_line > 0 || last_line < nb_lines - 1)) {

        if(first_line > 0) {
          first_line--;
          while(first_line > 0 && !match(&matcher, lines[first_line], 0, 0)) {
            first_line--;
          }
          if(match(&matcher, lines[first_line], 0, 0)) {
            nb_match++;
          }
        }

        if(nb_match < console_height - 1 && last_line < nb_lines - 1) {
          last_line++;
          while(last_line < nb_lines - 1 && !match(&matcher, lines[last_line], 0, 0)) {
            last_line++;
          }

          if(match(&matcher, lines[last_line], 0, 0)) {
            nb_match++;
          }
        }
      }

      /* Now we display them */

      for(l = first_line; l <= last_line; l++) {
        if(match(&matcher, lines[l], &nb_switches, switches)) {
          int k = 0;

          while(lines[l][k] && k < BUFFER_SIZE - 2 && k < console_width) {
            buffer[k] = lines[l][k];
            k++;
          }

          /* Highlight the highlighted line ... */

          if(l == new_focus_line) {
            if(show_long_lines && k >= console_width) {
              if(console_width >= 4) {
                buffer[console_width - 4] = ' ';
                buffer[console_width - 3] = '.';
                buffer[console_width - 2] = '.';
                buffer[console_width - 1] = '.';
              }
            } else {
              while(k < console_width) {
                buffer[k++] = ' ';
              }
            }
            attron(attr_focus_line);
            print_string_with_switches(buffer, k, console_width,
                                       nb_switches / 2, switches);
            attroff(attr_focus_line);
          } else {
            if(show_long_lines && k >= console_width) {
              if(console_width >= 4) {
                buffer[console_width - 4] = ' ';
                buffer[console_width - 3] = '.';
                buffer[console_width - 2] = '.';
                buffer[console_width - 1] = '.';
              }
            } else {
              buffer[k++] = '\n';
              buffer[k++] = '\0';
            }

            print_string_with_switches(buffer, k, console_width,
                                       nb_switches / 2, switches);
          }

          nb_printed_lines++;
        }
      }

      /* If we are on a focused line and we moved, this become the new
         focus line */

      if(motion != 0) {
        *current_focus_line = new_focus_line;
      }
    }

    *displayed_focus_line = new_focus_line;

    if(nb_printed_lines == 0) {
      attron(attr_error);
      addnstr("No selection", console_width);
      attroff(attr_error);
    }
  }

  /* Else, print a message saying that there are no lines to select from */

  else {
    attron(attr_error);
    addnstr("Empty choice", console_width);
    attroff(attr_error);
  }

  clrtobot();

  /* Draw the modeline */

  move(0, 0);

  attron(attr_modeline);

  for(k = 0; k < console_width; k++) buffer[k] = ' ';
  buffer[console_width] = '\0';
  addnstr(buffer, console_width);

  move(0, 0);

  /* There must be a more elegant way of moving the cursor at a
     location met during display */

  cursor_x = 0;

  if(title) {
    addstr(title);
    addstr(" ");
    cursor_x += strlen(title) + 1;
  }

  sprintf(buffer, "%d/%d ", nb_printed_lines, nb_lines);
  addstr(buffer);
  cursor_x += strlen(buffer);

  addnstr(pattern, cursor_position);
  cursor_x += cursor_position;

  if(pattern[cursor_position]) {
    addstr(pattern + cursor_position);
  } else {
    addstr(" ");
  }

  /* Add a few info about the mode we are in (regexp and/or case
     sensitive) */

  if(use_regexp || matcher.case_sensitive) {
    addstr(" [");
    if(use_regexp) {
      addstr("regexp");
    }

    if(matcher.case_sensitive) {
      if(use_regexp) {
        addstr(",");
      }
      addstr("case");
    }
    addstr("]");
  }

  move(0, cursor_x);

  attroff(attr_modeline);

  /* We are done */

  refresh();
  if(switches) { free(switches); }
  free_matcher(&matcher);
}

/*********************************************************************/

void store_line(struct hash_table_t *hash_table,
                const char *new_line,
                int *nb_lines, char **lines) {
  int dup;

  /* Remove the zsh history prefix */

  if(zsh_history && *new_line == ':') {
    while(*new_line && *new_line != ';') new_line++;
    if(*new_line == ';') new_line++;
  }

  /* Remove the bash history prefix */

  if(bash_history) {
    while(*new_line == ' ') new_line++;
    while(*new_line >= '0' && *new_line <= '9') new_line++;
    while(*new_line == ' ') new_line++;
  }

  /* Check for duplicates with the hash table and insert the line in
     the list if necessary */

  if(hash_table) {
    dup = add_and_get_previous_index(hash_table,
                                     new_line, *nb_lines, lines);
  } else {
    dup = -1;
  }

  if(dup < 0) {
    lines[*nb_lines] = safe_malloc((strlen(new_line) + 1) * sizeof(char));
    strcpy(lines[*nb_lines], new_line);
  } else {
    /* The string was already in there, so we do not allocate a new
       string but use the pointer to the first occurence of it */
    lines[*nb_lines] = lines[dup];
    lines[dup] = 0;
  }

  (*nb_lines)++;
}

void read_file(struct hash_table_t *hash_table,
               const char *input_filename,
               int nb_lines_max, int *nb_lines, char **lines) {

  char raw_line[BUFFER_SIZE];
  char *s;
  FILE *file;

  file = fopen(input_filename, "r");

  if(!file) {
    fprintf(stderr, "selector: Can not open `%s'.\n", input_filename);
    exit(EXIT_FAILURE);
  }

  while(*nb_lines < nb_lines_max && fgets(raw_line, BUFFER_SIZE, file)) {
    for(s = raw_line + strlen(raw_line) - 1; s > raw_line && *s == '\n'; s--) {
      *s = '\0';
    }
    store_line(hash_table, raw_line, nb_lines, lines);
  }

  fclose(file);
}

/*********************************************************************/

/* For long options that have no equivalent short option, use a
   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
enum
{
  OPT_BASH_MODE = CHAR_MAX + 1
};

static struct option long_options[] = {
  { "output-file", 1, 0, 'o' },
  { "pattern-separator", 1, 0, 's' },
  { "label-separator", 1, 0, 'x' },
  { "inject-in-tty", no_argument, 0, 'v' },
  { "add-control-qs", no_argument, 0, 'w' },
  { "monochrome", no_argument, 0, 'm' },
  { "no-beep", no_argument, 0, 'q' },
  { "revert-order", no_argument, 0, 'i' },
  { "remove-bash-prefix", no_argument, 0, 'b' },
  { "remove-zsh-prefix", no_argument, 0, 'z' },
  { "remove-duplicates", no_argument, 0, 'd' },
  { "regexp", no_argument, 0, 'e' },
  { "case-sensitive", no_argument, 0, 'a' },
  { "show-long-lines", no_argument, 0, 'j'},
  { "show-hits", no_argument, 0, 'j'},
  { "upper-case-makes-case-sensitive", no_argument, 0, 'u' },
  { "title", 1, 0, 't' },
  { "number-of-lines", 1, 0, 'l' },
  { "colors", 1, 0, 'c' },
  { "bash", no_argument, 0, OPT_BASH_MODE },
  { "help", no_argument, 0, 'h' },
  { 0, 0, 0, 0 }
};

int main(int argc, char **argv) {

  char output_filename[BUFFER_SIZE];
  char pattern[BUFFER_SIZE];
  int c, k, l, n;
  int cursor_position;
  int error = 0, show_help = 0, done = 0;
  int key;
  int current_focus_line, displayed_focus_line;

  int colors[4];
  int color_fg_modeline, color_bg_modeline;
  int color_fg_highlight, color_bg_highlight;

  char **lines, **labels;
  int nb_lines;
  struct hash_table_t *hash_table;
  char *bash_histsize;

  if(!isatty(STDIN_FILENO)) {
    fprintf(stderr, "selector: The standard input is not a tty.\n");
    exit(EXIT_FAILURE);
  }

  color_fg_modeline  = COLOR_WHITE;
  color_bg_modeline  = COLOR_BLACK;
  color_fg_highlight = COLOR_BLACK;
  color_bg_highlight = COLOR_YELLOW;

  setlocale(LC_ALL, "");

  strcpy(output_filename, "");

  while ((c = getopt_long(argc, argv, "o:s:x:vwmqf:ibzdeajyunt:l:c:-h",
                          long_options, NULL)) != -1) {

    switch(c) {

    case 'o':
      strncpy(output_filename, optarg, BUFFER_SIZE);
      break;

    case 's':
      pattern_separator = optarg[0];
      break;

    case 'x':
      label_separator = optarg[0];
      break;

    case 'v':
      output_to_vt_buffer = 1;
      break;

    case 'w':
      add_control_qs = 1;
      break;

    case 'm':
      with_colors = 0;
      break;

    case 'q':
      error_flash = 1;
      break;

    case 'i':
      inverse_order = 1;
      break;

    case 'b':
      bash_history = 1;
      break;

    case 'z':
      zsh_history = 1;
      break;

    case 'd':
      remove_duplicates = 1;
      break;

    case 'e':
      use_regexp = 1;
      break;

    case 'a':
      case_sensitive = 1;
      break;

    case 'j':
      show_long_lines = 1;
      break;

    case 'y':
      show_hits = 1;
      break;

    case 'u':
      upper_caps_makes_case_sensitive = 1;
      break;

    case 't':
      free(title);
      title = safe_malloc((strlen(optarg) + 1) * sizeof(char));
      strcpy(title, optarg);
      break;

    case 'l':
      str_to_positive_integers(optarg, &nb_lines_max, 1);
      break;

    case 'c':
      str_to_positive_integers(optarg, colors, 4);
      color_fg_modeline = colors[0];
      color_bg_modeline = colors[1];
      color_fg_highlight = colors[2];
      color_bg_highlight = colors[3];
      break;

    case 'h':
      show_help = 1;
      break;

    case OPT_BASH_MODE:
      /* Same as -c 7,4,0,3 -q */
      /* color_fg_modeline = 7; */
      /* color_bg_modeline = 4; */
      /* color_fg_highlight = 0; */
      /* color_bg_highlight = 3; */
      /* error_flash = 1; */
      /* Same as -b -i -d -v -w */
      bash_history = 1;
      inverse_order = 1;
      remove_duplicates = 1;
      output_to_vt_buffer = 1;
      add_control_qs = 1;
      bash_histsize = getenv("HISTSIZE");
      if(bash_histsize) {
        str_to_positive_integers(bash_histsize, &nb_lines_max, 1);
      }
      break;

    default:
      error = 1;
      break;
    }
  }

  if(error) {
    usage(stderr);
    exit(EXIT_FAILURE);
  }

  if(show_help) {
    usage(stdout);
    exit(EXIT_SUCCESS);
  }

  lines = safe_malloc(nb_lines_max * sizeof(char *));

  nb_lines = 0;

  if(remove_duplicates) {
    hash_table = new_hash_table(nb_lines_max * 10);
  } else {
    hash_table = 0;
  }

  while(optind < argc) {
    read_file(hash_table,
              argv[optind],
              nb_lines_max, &nb_lines, lines);
    optind++;
  }

  if(hash_table) {
    free_hash_table(hash_table);
  }

  /* Now remove the null strings */

  n = 0;
  for(k = 0; k < nb_lines; k++) {
    if(lines[k]) {
      lines[n++] = lines[k];
    }
  }

  nb_lines = n;

  if(inverse_order) {
    for(l = 0; l < nb_lines / 2; l++) {
      char *s = lines[nb_lines - 1 - l];
      lines[nb_lines - 1 - l] = lines[l];
      lines[l] = s;
    }
  }

  /* Build the labels from the strings, take only the part before the
     label_separator and transform control characters to printable
     ones */

  labels = safe_malloc(nb_lines * sizeof(char *));

  for(l = 0; l < nb_lines; l++) {
    char *s, *t;
    int e = 0;
    const char *u;
    t = lines[l];

    while(*t && *t != label_separator) {
      u = unctrl(*t++);
      e += strlen(u);
    }

    labels[l] = safe_malloc((e + 1) * sizeof(char));
    t = lines[l];
    s = labels[l];
    while(*t && *t != label_separator) {
      u = unctrl(*t++);
      while(*u) { *s++ = *u++; }
    }
    *s = '\0';
  }

  pattern[0] = '\0';

  cursor_position = 0;

  /* Here we start to display with curse */

  initscr();
  cbreak();
  noecho();
  intrflush(stdscr, FALSE);

  /* So that the arrow keys work */
  keypad(stdscr, TRUE);

  attr_error = A_STANDOUT;
  attr_modeline = A_REVERSE;
  attr_focus_line = A_STANDOUT;
  attr_hits = A_BOLD;

  if(with_colors && has_colors()) {

    start_color();

    if(color_fg_modeline < 0  || color_fg_modeline >= COLORS ||
       color_bg_modeline < 0  || color_bg_modeline >= COLORS ||
       color_fg_highlight < 0 || color_bg_highlight >= COLORS ||
       color_bg_highlight < 0 || color_bg_highlight >= COLORS) {
      echo();
      endwin();
      fprintf(stderr, "selector: Color numbers have to be between 0 and %d.\n",
              COLORS - 1);
      exit(EXIT_FAILURE);
    }

    init_pair(1, color_fg_modeline, color_bg_modeline);
    attr_modeline = COLOR_PAIR(1);

    init_pair(2, color_fg_highlight, color_bg_highlight);
    attr_focus_line = COLOR_PAIR(2);

    init_pair(3, COLOR_WHITE, COLOR_RED);
    attr_error = COLOR_PAIR(3);

  }

  current_focus_line = 0;
  displayed_focus_line = 0;

  update_screen(&current_focus_line, &displayed_focus_line,
                0,
                nb_lines, labels, cursor_position, pattern);

  do {
    int motion = 0;

    key = getch();

    if(key >= ' ' && key <= '~') { /* Insert character */
      insert_char(pattern, &cursor_position, key);
    }

    else if(key == KEY_BACKSPACE ||
            key == '\010' || /* ^H */
            key == '\177') { /* ^? */
      backspace_char(pattern, &cursor_position);
    }

    else if(key == KEY_DC ||
            key == '\004') { /* ^D */
      delete_char(pattern, &cursor_position);
    }

    else if(key == KEY_HOME) {
      current_focus_line = 0;
    }

    else if(key == KEY_END) {
      current_focus_line = nb_lines - 1;
    }

    else if(key == KEY_NPAGE) {
      motion = 10;
    }

    else if(key == KEY_PPAGE) {
      motion = -10;
    }

    else if(key == KEY_DOWN ||
            key == '\016') { /* ^N */
      motion = 1;
    }

    else if(key == KEY_UP ||
            key == '\020') { /* ^P */
      motion = -1;
    }

    else if(key == KEY_LEFT ||
            key == '\002') { /* ^B */
      if(cursor_position > 0) cursor_position--;
      else error_feedback();
    }

    else if(key == KEY_RIGHT ||
            key == '\006') { /* ^F */
      if(pattern[cursor_position]) cursor_position++;
      else error_feedback();
    }

    else if(key == '\001') { /* ^A */
      cursor_position = 0;
    }

    else if(key == '\005') { /* ^E */
      cursor_position = strlen(pattern);
    }

    else if(key == '\022') { /* ^R */
      use_regexp = !use_regexp;
    }

    else if(key == '\011') { /* ^I */
      case_sensitive = !case_sensitive;
    }

    else if(key == '\025') { /* ^U */
      kill_before_cursor(pattern, &cursor_position);
    }

    else if(key == '\013') { /* ^K */
      kill_after_cursor(pattern, &cursor_position);
    }

    else if(key == '\014') { /* ^L */
      /* I suspect that we may sometime mess up the display, so ^L is
         here to force a full refresh */
      clear();
    }

    else if(key == '\007' || /* ^G */
            key == '\033' || /* ^[ (escape) */
            key == '\n' ||
            key == KEY_ENTER) {
      done = 1;
    }

    else if(key == KEY_RESIZE || key == -1) {
      /* Do nothing when the tty is resized */
    }

    else {
      /* Unknown key */
      error_feedback();
    }

    update_screen(&current_focus_line, &displayed_focus_line,
                  motion,
                  nb_lines, labels, cursor_position, pattern);

  } while(!done);

  echo();
  endwin();

  /* Here we come back to standard display */

  if(key == KEY_ENTER || key == '\n') {

    char *t;

    if(displayed_focus_line >= 0 && displayed_focus_line < nb_lines) {
      t = lines[displayed_focus_line];
      if(label_separator) {
        while(*t && *t != label_separator) t++;
        if(*t) t++;
      }
    } else {
      t = 0;
    }

    if(output_to_vt_buffer && t) {
      inject_into_tty_buffer(t, add_control_qs);
    }

    if(output_filename[0]) {
      FILE *out = fopen(output_filename, "w");
      if(out) {
        if(t) {
          fprintf(out, "%s", t);
        }
        fprintf(out, "\n");
      } else {
        fprintf(stderr,
                "selector: Can not open %s for writing.\n",
                output_filename);
        exit(EXIT_FAILURE);
      }
      fclose(out);
    }

  } else {
    printf("Aborted.\n");
  }

  for(l = 0; l < nb_lines; l++) {
    free(lines[l]);
    free(labels[l]);
  }

  free(labels);
  free(lines);
  free(title);

  exit(EXIT_SUCCESS);
}
